﻿Imports System
Imports System.Threading
Imports System.Diagnostics

Namespace MsdnMag.Exchange.PushNotificationClient
    ''' <summary>
    ''' Class for managing the retry logic around operations.
    ''' </summary>
    Public NotInheritable Class RetryManager
        Private Const DEFAULT_RETRIES_COUNT As Integer = 3
        Private Const WAIT_TIME_MILLISECONDS As Integer = 1000

        ''' <summary>
        ''' Invoke the method with no parameters and a default retry limit.
        ''' <param name="action"></param>
        ''' </summary>
        Private Sub New()
        End Sub
        Public Shared Sub InvokeWithRetries(ByVal action As Action)
            InvokeWithRetries(action, DEFAULT_RETRIES_COUNT)
        End Sub

        ''' <summary>
        ''' Invoke the method with no parameters and the specified retry limit.
        ''' </summary>
        ''' <param name="action"></param>
        ''' <param name="numberOfRetries"></param>
        Public Shared Sub InvokeWithRetries(ByVal action As Action, ByVal numberOfRetries As Integer)
            Dim currentRetriesCount As Integer = 0
            Do While currentRetriesCount < numberOfRetries
                currentRetriesCount += 1

                Try
                    action.Invoke()

                    Return
                Catch ex As Exception
                    If currentRetriesCount < numberOfRetries Then
                        Trace.WriteLine(String.Format("InvokeWithRetries: Error: {0}. Current attempt: {1}. Sleeping for {2} milliseconds before retrying...", ex.Message, currentRetriesCount, WAIT_TIME_MILLISECONDS))

                        Thread.Sleep(WAIT_TIME_MILLISECONDS)

                        Continue Do
                    Else
                        Trace.WriteLine(String.Format("InvokeWithRetries: Error: {0}. Current attempt: {1}. Retries expired. Rethrowing exception...", ex.Message, currentRetriesCount, WAIT_TIME_MILLISECONDS))

                        Throw
                    End If
                End Try
            Loop
        End Sub
    End Class
End Namespace
