﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Generic
Imports System.Text
Imports Microsoft.TeamFoundation.WorkItemTracking.Client
Imports Microsoft.Build.Utilities
Imports Microsoft.Build.Framework
Imports System.Collections
Imports System.Globalization
Imports Microsoft.TeamFoundation.Client
Imports Microsoft.TeamFoundation.VersionControl.Client

Namespace MsdnMag.Exchange.PushNotificationClient
    ''' <summary>
    ''' Updates the specified fields of a given work item.
    ''' </summary>
    Public Class UpdateWorkItem
        Inherits Task
#Region "Private Fields"

        Private _properties As String
        Private _workItemTeamProject As String
        Private _tfsUrl As String
        Private _workItemId As Integer
        Private _history As String
        Private _numErrors As Integer
        Private _tfs As TfsTeamProjectCollection
        Private _versionControlServer As VersionControlServer
        Private _workItemStore As WorkItemStore
        Private _associatedWorkItems As String
        Private _associatedChangesets As String

#End Region

#Region "Methods"

        Public Overrides Function Execute() As Boolean
            Dim workItem As WorkItem = Me.WorkItemStore.GetWorkItem(_workItemId)

            Me.UpdateUserSpecifiedFields(workItem, PropertyParser.GetTable(Me.WorkItemFieldValues))

            If Not String.IsNullOrWhiteSpace(Me.AssociatedWorkItems) Then
                Me.AddLinkToAssociatedWorkItems(workItem, Me.AssociatedWorkItems)
            End If

            If Not String.IsNullOrWhiteSpace(Me.AssociatedChangesets) Then
                Me.AddLinkToAssociatedChangesets(workItem, Me.AssociatedChangesets)
            End If

            ' Update the history:
            If Not String.IsNullOrWhiteSpace(_history) Then
                workItem.History = _history
            End If

            Me.SaveWorkItem(workItem)

            Log.LogMessage("UpdateWorkItem: Workitem Updated: Id: {0}.", New Object() {workItem.Id})

            Return True
        End Function

        Private Sub AddLinkToAssociatedChangesets(ByVal workItem As WorkItem, ByVal associatedChangesets As String)
            Dim associatedChangesetsArray() As String = associatedChangesets.Split(New Char() {","c}, StringSplitOptions.RemoveEmptyEntries)

            If associatedChangesetsArray IsNot Nothing AndAlso associatedChangesetsArray.Length > 0 Then
                For Each item In associatedChangesetsArray
                    Try
                        Dim changeset As Changeset = Me.VersionControlServer.GetChangeset(Convert.ToInt32(item))
                        Dim link As Link = New ExternalLink(Me.WorkItemStore.RegisteredLinkTypes("Source Code File"), changeset.ArtifactUri.AbsoluteUri)
                        workItem.Links.Add(link)

                        Log.LogMessage(String.Format("UpdateWorkItem: Added a link to associated changeset: {0}.", New Object() {changeset.ChangesetId}))
                    Catch ex As Exception
                        Log.LogWarning(String.Format("Unable to associate changeset: {0}." + vbCrLf + "Error Message: {1}" + vbCrLf + "StackTrace: {2}.", New Object() {item, ex.Message, ex.StackTrace}))
                    End Try
                Next item
            Else
                Log.LogWarning("UpdateWorkItem: AssociatedWorkItems list is empty.")
                Return
            End If
        End Sub

        Private Sub AddLinkToAssociatedWorkItems(ByVal workItem As WorkItem, ByVal associatedWorkItems As String)
            Dim associatedWorkItemsArray() As String = associatedWorkItems.Split(New Char() {","c}, StringSplitOptions.RemoveEmptyEntries)

            If associatedWorkItemsArray IsNot Nothing AndAlso associatedWorkItemsArray.Length > 0 Then
                For Each item In associatedWorkItemsArray
                    Try
                        Dim workItemId As Integer = Convert.ToInt32(item)

                        If Not IsWorkItemAlreadyLinked(workItemId, workItem.Links) Then
                            Dim link As New RelatedLink(workItemId)
                            workItem.Links.Add(link)

                            Log.LogMessage(String.Format("UpdateWorkItem: Added a link to associated work item: {0}.", New Object() {workItemId}))
                        Else
                            Log.LogMessage(String.Format("UpdateWorkItem: Skipped linking to associated work item as it is already linked: {0}.", New Object() {workItemId}))
                        End If
                    Catch ex As Exception
                        Log.LogWarning(String.Format("Unable to associate work item: {0}." & vbCrLf & "Error Message: {1}" & vbCrLf & "StackTrace: {2}.", New Object() {item, ex.Message, ex.StackTrace}))
                    End Try
                Next item
            Else
                Log.LogWarning("UpdateWorkItem: AssociatedWorkItems list is empty.")
                Return
            End If
        End Sub

        Friend Shared Function IsWorkItemAlreadyLinked(ByVal workItemId As Integer, ByVal links As LinkCollection) As Boolean
            If workItemId = 0 OrElse links Is Nothing OrElse links.Count = 0 Then
                Return False
            End If

            For Each link In links
                If TypeOf link Is RelatedLink AndAlso (CType(link, RelatedLink)).RelatedWorkItemId = workItemId Then
                    Return True
                End If
            Next link
            Return False
        End Function

        Private Sub SaveWorkItem(ByVal workItem As WorkItem)
            Try
                workItem.Save()
                Me.WorkItemId = workItem.Id
            Catch exception As Exception
                Dim list As ArrayList = workItem.Validate()
                If list.Count > 0 Then
                    Dim builder As New StringBuilder()
                    For i = 0 To list.Count - 1
                        Dim field As Field = CType(list(i), Field)
                        If i <> 0 Then
                            builder.Append(CultureInfo.CurrentCulture.TextInfo.ListSeparator)
                        End If
                        builder.Append(String.Format("Field Error: {0} : {1}.", New Object() {field.Name, field.Value}))
                    Next i
                    Throw New Exception(String.Format("Field Errors Saving New Work Item: {0}.", New Object() {builder.ToString()}), exception)
                End If
                If Not workItem.Project.HasWorkItemWriteRights Then
                    Throw New Exception("No Permission To Write WI.", exception)
                End If
                Throw
            End Try
        End Sub

        Private Sub SetFieldIfExists(ByVal workItem As WorkItem, ByVal fieldRefName As String, ByVal value As String)
            If workItem.Type.FieldDefinitions.Contains(fieldRefName) Then
                Dim field As Field = workItem.Fields(fieldRefName)
                If field.IsEditable AndAlso ((Not field.IsLimitedToAllowedValues) OrElse field.AllowedValues.Contains(value)) Then
                    workItem(fieldRefName) = value
                End If
            End If
        End Sub

        Private Sub UpdateUserSpecifiedFields(ByVal workItem As WorkItem, ByVal propertiesTable As Dictionary(Of String, String))
            If propertiesTable IsNot Nothing Then
                For Each pair As KeyValuePair(Of String, String) In propertiesTable
                    Try
                        Dim field As Field = workItem.Fields(pair.Key)
                        field.Value = pair.Value
                    Catch exception As Exception
                        Log.LogWarning(String.Format("Unable To Update WI Field: {0} - {1}.", New Object() {pair.Key, exception.Message}))
                        Continue For
                    End Try
                Next pair
            End If
        End Sub

        Private Sub NonFatalError(ByVal sender As Object, ByVal e As ExceptionEventArgs)
            Dim errorMessage As String = Nothing
            If e.Exception IsNot Nothing Then
                errorMessage = e.Exception.Message
            End If
            If e.Failure IsNot Nothing Then
                If e.Failure.Severity = SeverityType.Warning Then
                    Return
                End If
                errorMessage = e.Failure.Message
            End If
            Log.LogError(errorMessage)
            Me._numErrors += 1
        End Sub

#End Region

#Region "Properties"

        <Required()>
        Public Property WorkItemTeamProject() As String
            Get
                Return Me._workItemTeamProject
            End Get
            Set(ByVal value As String)
                Me._workItemTeamProject = value
            End Set
        End Property

        Public Property AssociatedWorkItems() As String
            Get
                Return _associatedWorkItems
            End Get
            Set(ByVal value As String)
                _associatedWorkItems = value
            End Set
        End Property

        Public Property AssociatedChangesets() As String
            Get
                Return _associatedChangesets
            End Get
            Set(ByVal value As String)
                _associatedChangesets = value
            End Set
        End Property

        Public Property WorkItemFieldValues() As String
            Get
                Return Me._properties
            End Get
            Set(ByVal value As String)
                Me._properties = value
            End Set
        End Property

        Public Property History() As String
            Get
                Return _history
            End Get
            Set(ByVal value As String)
                _history = value
            End Set
        End Property

        Friend ReadOnly Property WorkItemStore() As WorkItemStore
            Get
                If Me._workItemStore Is Nothing Then
                    _workItemStore = CType(Me.Tfs.GetService(GetType(WorkItemStore)), WorkItemStore)
                End If
                Return _workItemStore
            End Get
        End Property

        <Required()>
        Public Property TeamFoundationServerUrl() As String
            Get
                Return Me._tfsUrl
            End Get
            Set(ByVal value As String)
                Me._tfsUrl = value
            End Set
        End Property

        Protected ReadOnly Property Tfs() As TfsTeamProjectCollection
            Get
                If Me._tfs Is Nothing Then
                    Me._tfs = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(New Uri(Me.TeamFoundationServerUrl))
                End If
                Return Me._tfs
            End Get
        End Property

        Friend ReadOnly Property VersionControlServer() As VersionControlServer
            Get
                If Me._versionControlServer Is Nothing Then
                    Me._versionControlServer = CType(Me.Tfs.GetService(GetType(VersionControlServer)), VersionControlServer)
                    AddHandler _versionControlServer.NonFatalError, AddressOf NonFatalError
                End If
                Return Me._versionControlServer
            End Get
        End Property

        <Required()>
        Public Property WorkItemId() As Integer
            Get
                Return Me._workItemId
            End Get
            Set(ByVal value As Integer)
                Me._workItemId = value
            End Set
        End Property

#End Region
    End Class
End Namespace
